#! /usr/local/bin/perl

##
## MAKE_GEMM.pl
##
## Perl script to generate the PHiPAC GEMM library.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##      Jeff Bilmes <bilmes@icsi.berkeley.edu>
##
## $Header: /db/bilmes/phipac/dist/PHiPAC-beta/PHiPAC/search-2.2/RCS/make_gemm.pl,v 1.2 1998/02/20 01:29:59 bilmes Exp $
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);

   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$search_mode = "";
$useage = "$0 <machine specification file> <single|double> <information file>";

## Read search parameters.
require "search_readparams.pl";
&read_machine_specs($ARGV[0]);
$prec = $ARGV[1];
$info_file = $ARGV[2];

$gemm = "";

if ($prec eq "single")
{
   $gemm = "sgemm";
}
elsif ($prec eq "double")
{
   $gemm = "dgemm";
}
else
{
   die "***Precision must be \"single\" or \"double\".\n\n";
}
   

## Load supporting routines.
require "search_supp.pl";


if ((-e "$working_dir/PHiPAC_$gemm") &&
    (-d "$working_dir/PHiPAC_$gemm"))
{
   # PHiPAC GEMM directory already exist,
   warn "***PHiPAC GEMM directory PHiPAC_$gemm already exists, removing.\n\n";
   $cc_call = "rm -rf $working_dir/PHiPAC_$gemm;";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;
   system($cc_call);
   die "***Can't remove old PHiPAC GEMM directory (exit status $?)\n"
      if ($? != 0);
}

chop($hostname = `hostname`);
chop($date = `date`);
print STDOUT "\n\nMake_GEMM started on $hostname on $date\n\n";
&flush(STDOUT);

## Change to temporary directory
chdir($tmp_dir);

$cc_call = "rm -rf src Makefile; mkdir src";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
system($cc_call);
die "***Can't create GEMM source code directory (exit status $?)\n"
   if ($? != 0);

print STDOUT "Generating PHiPAC GEMM routines...\n\n"
   if $verbosity > 0;

## Create cases.
&create_cases;

@obj_files_sgemm = ();
@obj_files_dgemm = ();

# open and initialize Makefile
open(MAKE_FILE,">$tmp_dir/Makefile")
   || die "***Can't open $tmp_dir/Makefile.\n";
printf(MAKE_FILE &file_header_str);

printf(MAKE_FILE &makefile_header);

##
## Loop through all cases
## and generate PHiPAC code for each case
##

CASE:
foreach $case (sort(keys(%cases)))
{
   print STDOUT "Generating $case...\n\n"
      if $verbosity > 0;

   @obj_files_case = ();
   ($prec, $matop, $alphatype) = split(/\./, $case);

   $name = $prec . "_" . $matop . "_" . $alphatype;
   $name_general = $name . "_general";
   $name_nofringe = $name . "_nofringe";
   $name_fixed = $name . "_fixed";
   $name_l1 = $name . "_l1";

   # select the blocking information for this case
   ($level, $l2, $l0gen, $l1, $l0core, $cflags)
      = split(/\;/, $cases{$case});
   ($m2,$k2,$n2) = split(/\,/, $l2);
   ($m0g,$k0g,$n0g,$l0gspopt) = split(/\,/, $l0gen);
   ($m1,$k1,$n1,$l1codeuse) = split(/\,/, $l1);
   ($m0,$k0,$n0,$l0spopt) = split(/\,/, $l0core);
       
   if ($level == 0)
   {
      # Use L0 blocking only

      # generate general core code
      $status = &generate_core("-file ./src/mm_$name.c -routine_name mm_$name -beta c",
			       $prec, $matop, $alphatype,
			       "general", $l0gspopt,
			       $m0g,$k0g,$n0g);
      die "***Can't create Makefile.\n" if ($status != 0);

      # compile general core code
      &makefile_compile_code("mm_$name",$cflags);

      push(@obj_files_case, "./obj/mm_$name.o");;
   }
   else
   {
      # generate general core code
      $status = &generate_core("-file ./src/mm_$name_general.c -routine_name mm_$name_general",
			       $prec, $matop, $alphatype,
			       "general", $l0gspopt,
			       $m0g,$k0g,$n0g);
      die "***Can't create Makefile.\n" if ($status != 0);

      # compile general core code
      &makefile_compile_code("mm_$name_general",$cflags);

      push(@obj_files_case, " ./obj/mm_$name_general.o");
      $generator_controls = " -gen_rout mm_$name_general";

      if (($l1codeuse eq "fixed_nofringe_general") ||
	  ($l1codeuse eq "nofringe_general"))
      {
	 # generate nofringe core code
	 $status = &generate_core("-file ./src/mm_$name_nofringe.c -routine_name mm_$name_nofringe",
				  $prec, $matop, $alphatype,
				  "nofringe", $l0spopt,
				  $m0,$k0,$n0);
	 die "***Can't create Makefile.\n" if ($status != 0);

	 # compile nofringe core code
	 &makefile_compile_code("mm_$name_nofringe",$cflags);
	 
	 push(@obj_files_case, " ./obj/mm_$name_nofringe.o");
	 $generator_controls .= " -gen_nf_rout mm_$name_nofringe";
      }

      if ($l1codeuse eq "fixed_nofringe_general")
      {
	 # also generate fixed core code
	 $status = &generate_core("-file ./src/mm_$name_fixed.c -routine_name mm_$name_fixed",
				  $prec, $matop, $alphatype,
				  "fixed",$l0spopt,
				  $m0,$k0,$n0,
				  $m1,$k1,$n1);
	 die "***Can't create Makefile.\n" if ($status != 0);

	 # also compile fixed core code
	 &makefile_compile_code("mm_$name_fixed", $cflags);
	 
	 push(@obj_files_case, " ./obj/mm_$name_fixed.o");
	 $generator_controls .= " -fixed_rout mm_$name_fixed";
      }

      if ($level == 1)
      {
	 $generator_controls .= " -file ./src/mm_$name.c -routine_name mm_$name -beta c";

	 $status = &generate_level($generator_controls,
				   $prec, $matop, $alphatype,
				   $m1,$k1,$n1,
				   $m0, $k0, $n0,
				   $l0spopt);
	 die "***Can't create Makefile.\n" if ($status != 0);
	 
	 # compile generated source code
	 &makefile_compile_code("mm_$name", $cflags);
	 
	 push(@obj_files_case, " ./obj/mm_$name.o");
      }
      elsif ($level == 2)
      {
	 $generator_controls .= " -file ./src/mm_$name_l1.c -routine_name mm_$name_l1";

	 # generate L1 code
	 $status = &generate_level($generator_controls,
				   $prec, $matop, $alphatype,
				   $m1,$k1,$n1,
				   $m0, $k0, $n0,
				   $l0spopt);
	 die "***Can't create Makefile.\n" if ($status != 0);

	 # compile L1 code
	 &makefile_compile_code("mm_$name_l1", $cflags);

	 push(@obj_files_case, " ./obj/mm_$name_l1.o");

	 # generate L2 code
	 $spadd = substr($l0spopt,0,1) - 1;
	 $mm = $m0 * $m1 * $m2;
	 $kk = ($k0 * $k1 + $spadd) * $k2;
	 $nn = $n0 * $n1 * $n2;
	 $generator_controls = " -gen_rout mm_$name_l1";
	 $generator_controls .= " -file ./src/mm_$name.c -routine_name mm_$name -beta c";
	 
	 $status = &generate_level($generator_controls,
				   $prec, $matop, $alphatype,
				   $mm,$kk,$nn,
				   1, 1, 1,
				   "");
	 die "***Can't create Makefile.\n" if ($status != 0);

	 # compile L2 code
	 &makefile_compile_code("mm_$name",$cflags);

	 push(@obj_files_case, " ./obj/mm_$name.o");
      }
      else
      {
	 die "***Blocking level $level not supported.\n";
      }
   }

   if ($prec eq "single")
   {
      push(@obj_files_sgemm, @obj_files_case);
   }
   elsif ($prec eq "double")
   {
      push(@obj_files_dgemm, @obj_files_case);
   }

} # end loop for CASE


##
## Compile GEMM interfaces
##

@obj_files = ();

if ($prec eq "single")
{
   @obj_files = @obj_files_sgemm;
}

if ($prec eq "double")
{
   @obj_files = @obj_files_dgemm;
}

$cc_call = "cp $path_PHiPAC/$dir_SEARCH/phipac_$gemm.c ./src/";
system($cc_call);
die "***Can't copy PHiPAC GEMM interface source codes (exit status $?)\n"
   if ($? != 0);

&makefile_compile_code("phipac_$gemm", "-O");

printf(MAKE_FILE "lib$gemm.a: ./obj ./obj/phipac_$gemm.o ");
printf(MAKE_FILE join(" ", @obj_files));
printf(MAKE_FILE "\n");
printf(MAKE_FILE "\tar rcv ./lib$gemm.a ./obj/phipac_$gemm.o ");
printf(MAKE_FILE join(" ", @obj_files));
printf(MAKE_FILE "\n");
printf(MAKE_FILE "\tranlib ./lib$gemm.a\n\n");


# close Makefile
printf(MAKE_FILE "\n\n");
close(MAKE_FILE);


##
## Termination
##

## Change back to working directory
chdir($working_dir);

$cc_call = "rm -rf $working_dir/PHiPAC_$gemm; mkdir $working_dir/PHiPAC_$gemm";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
system($cc_call);
die "***Can't create GEMM directory (exit status $?)\n"
   if ($? != 0);

$cc_call = "cp -r $tmp_dir/src $tmp_dir/Makefile $working_dir/PHiPAC_$gemm";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
system($cc_call);
die "***Can't copy source code to GEMM directory (exit status $?)\n"
   if ($? != 0);

print STDOUT "\n\nPHiPAC_gemm directory PHiPAC_$gemm generated.\n";
print STDOUT "\n\nTo compile GEMM library, type:\n";
print STDOUT "\tcd PHiPAC_$gemm; make\n";
print STDOUT "\n\nTo generate tar-file for distribution, type:\n";
print STDOUT "\tcd PHiPAC_$gemm; make tar\n";


chop($date = `date`);
print STDOUT "\n\nMake_GEMM ended on $hostname at $date\n\n";

exit(0);

##
## END
##




##
## sub create_cases
##
## Generate all cases of routines to be compiled
##
## The returned value of this subroutine is the total number of
## cases generated.
##
sub create_cases
{
   local($key);
   local($level);

   ##
   ## create array %cases
   ##

   %cases = ();

   foreach $prec ("single", "double")
   {
      foreach $matop ("NN", "NT", "TN", "TT")
      {
	 foreach $alphatype ("1", "c")
	 {
	    $cases{"$prec.$matop.$alphatype"} = "";
	 }
      }
   }

   $has_single = 0;
   $has_double = 0;

   open(TOP_FILE, "$working_dir/$info_file")
      || die "***Can't open $working_dir/$info_file.\n";

   CASE:
   while(<TOP_FILE>)
   {
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      if (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 $_ =~ /^(\w+)\.(\w+)\.(\w+):\s+([0-9.]+)$/;
	 # the fields are:
	 #   (1) precision, (2) matrix operation, (3) alpha type,
	 #   (4) performance
	 
	 $key = "$1.$2.$3";

	 $has_single = 1 if ($1 eq "single");
	 $has_double = 1 if ($1 eq "double");

	 local($l2,$l0gen,$l1,$l0core,$cflags);
	 local($level);

	 ($l2,$l0gen,$l1,$l0core,$cflags) = ("","","","","");
	 $level = -1;

	 $_ = <TOP_FILE>;
	 while (substr($_,0,1) eq "#")
	 {
	    $_ =~ /^\# (.+) = (.*)$/;

	    if ($1 eq "L2")
	    {
	       $l2 = $2;
	       $level = 2 if ($level < 2);
	    }
	    elsif ($1 eq "L0GEN")
	    {
	       $l0gen = $2;
	       $level = 1 if ($level < 1);
	    }
	    elsif ($1 eq "L1")
	    {
	       $l1 = $2;
	       $level = 1 if ($level < 1);
	    }
	    elsif ($1 eq "L0CORE")
	    {
	       $l0core = $2;
	       $level = 0 if ($level < 0);
	    }
	    elsif ($1 eq "cflags")
	    {
	       $cflags = $2;
	    }
	    else
	    {
	       warn "***Ignoring information $_.\n";
	    }

	    $_ = <TOP_FILE>;
	 }

	 if ($level < 0)
	 {
	    warn "***No blocking information available for case $key.\n";
	    next CASE;
	 }

	 
	 ## compromises for incomplete information

	 if ($l0gen eq "")
	 {
	    warn "***Compromise: using L0CORE information for L0GEN in case $key.\n";
	    $l0gen = $l0core;
	 }

	 $cases{$key} = "$level;$l2;$l0gen;$l1;$l0core;$cflags";
      }
   } # end loop for CASE

   close(TOP_FILE);


   ## compromise for incomplete information

   foreach $prec ("single", "double")
   {
      foreach $matop ("NN", "NT", "TN", "TT")
      {
	 if (($cases{"$prec.$matop.1"} ne "") &&
	     ($cases{"$prec.$matop.c"} eq ""))
	 {
	    warn "***Compromise: using $prec.$matop.1 information for $prec.$matop.c\n";
	    $cases{"$prec.$matop.c"} = $cases{"$prec.$matop.1"};
	 }
	 elsif (($cases{"$prec.$matop.c"} ne "") &&
		($cases{"$prec.$matop.1"} eq ""))
	 {
	    warn "***Compromise: using $prec.$matop.c information for $prec.$matop.1\n";
	    $cases{"$prec.$matop.1"} = $cases{"$prec.$matop.c"};
	 }
      }

      if (($cases{"$prec.NN.1"} ne "") &&
	  ($cases{"$prec.NN.c"} ne "") &&
	  ($cases{"$prec.TT.1"} eq "") &&
	  ($cases{"$prec.TT.c"} eq ""))
      {
	 warn "***Compromise: using $prec.NN.* information for $prec.TT.*\n";
	 $cases{"$prec.TT.1"} = $cases{"$prec.NN.1"};
	 $cases{"$prec.TT.c"} = $cases{"$prec.NN.c"};
      }
      elsif (($cases{"$prec.TT.1"} ne "") &&
	     ($cases{"$prec.TT.c"} ne "") &&
	     ($cases{"$prec.NN.1"} eq "") &&
	     ($cases{"$prec.NN.c"} eq ""))
      {
	 warn "***Compromise: using $prec.TT.* information for $prec.NN.*\n";
	 $cases{"$prec.NN.1"} = $cases{"$prec.TT.1"};
	 $cases{"$prec.NN.c"} = $cases{"$prec.TT.c"};
      }
   }

   local($info_OK);

   $info_OK = 1;

   foreach $prec ("single", "double")
   {
      foreach $matop ("NN", "NT", "TN", "TT")
      {
	 foreach $alphatype ("1", "c")
	 {
	    if (($prec eq "single" && !$has_single) ||
		($prec eq "double" && !$has_double))
	    {
	       delete $cases{"$prec.$matop.$alphatype"};
	    }
	    elsif ($cases{"$prec.$matop.$alphatype"} eq "")
	    {
	       $info_OK = 0;
	       warn "***Incomplete information for case $prec.$matop.$alphatype\n";
	    }
	 }
      }
   }
	  
   die "***Can't create Makefile with incomplete information.\n"
      if (!$info_OK);

   print STDOUT "\n\n"; &flush(STDOUT);
}




##
## sub file_header_str
##
## Returns a file header string to be printed at the beginning of
## the Makefile.
##
sub file_header_str
{
   local($output_str);

   $output_str = "";
   
   $output_str .= "##\n";
   $output_str .= "## PHiPAC GEMM Makefile.\n";
   $output_str .= "##\n";

   $output_str .= "## Automatically generated by MAKE_GEMM.PL on $hostname on $date\n";
   $output_str .= "## using the information in:\n";
   $output_str .= "##   $working_dir/$info_file\n";
   $output_str .= "##\n";

   $output_str .= "## Copyright (c) 1997 The Regents of the University of California.  All\n";
   $output_str .= "## rights reserved.  Permission to use, copy, modify, and distribute\n";
   $output_str .= "## this software and its documentation for any purpose, without fee, and\n";
   $output_str .= "## without written agreement is hereby granted, provided that the above\n";
   $output_str .= "## copyright notice and the following two paragraphs appear in all copies\n";
   $output_str .= "## of this software.\n";
   $output_str .= "##\n";
   $output_str .= "## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY\n";
   $output_str .= "## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES\n";
   $output_str .= "## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF\n";
   $output_str .= "## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF\n";
   $output_str .= "## SUCH DAMAGE.\n";
   $output_str .= "##\n";
   $output_str .= "## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n";
   $output_str .= "## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n";
   $output_str .= "## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE\n";
   $output_str .= "## PROVIDED HEREUNDER IS ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF\n";
   $output_str .= "## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,\n";
   $output_str .= "## ENHANCEMENTS, OR MODIFICATIONS.\n";
   $output_str .= "##\n";

   $output_str;			# return value
}


##
## sub makefile_header
##
## Returns a string of common Makefile initializations to be printed
## at the beginning of the Makefile.
##
sub makefile_header
{
   local($output_str);

   $output_str = "\n\n";
   
   $output_str .= "all:";
   if ($prec eq "single")
   {
      $output_str .= " libsgemm.a";
   }
   if ($prec eq "double")
   {
      $output_str .= " libdgemm.a";
   }
   $output_str .= "\n\n";

   $output_str .= "./obj:\n";
   $output_str .= "\trm -rf obj\n";
   $output_str .= "\tmkdir obj\n\n";

   $output_str .= "tar: all Makefile src\n";
   $output_str .= "\ttar -cf PHiPAC_$gemm.tar Makefile src";
   if ($prec eq "single")
   {
      $output_str .= " libsgemm.a";
   }
   if ($prec eq "double")
   {
      $output_str .= " libdgemm.a";
   }
   $output_str .= "\n\n";

   $output_str .= "clean:\n";
   $output_str .= "\trm -rf obj lib?gemm.a PHiPAC_$gemm.tar\n\n";

   $output_str;			# return value
}


##
## sub makefile_compile_code
##
## Writes instructions to the Makefile to compile the generated code.
##
sub makefile_compile_code
{
   local($name, $cflags) = @_;
   local($cc_call);

   $cc_call = "$cc $cflags -o ./obj/$name.o -c ./src/$name.c";

   printf(MAKE_FILE "./obj/$name.o: ./src/$name.c\n");
   printf(MAKE_FILE "\t$cc_call\n\n");
}



