#! /usr/local/bin/perl

##
## SEARCH.PL
##
## Perl script to search for the optimal blocking sizes
## for matrix-matrix multiplication code generated by MM_GEN.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##      Jeff Bilmes <bilmes@icsi.berkeley.edu>
##
## $Header: /db/bilmes/phipac/dist/PHiPAC-beta/PHiPAC/search-2.2/RCS/search.pl,v 1.1 1997/10/09 17:57:38 bilmes Exp $
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);

   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$search_mode = "";
$useage = "$0 [-long|-default|-short] -machine <machine specification file> -prec <single|double> -ccopt <compiler options file> -level <levels of blocking>\n";

# parse command line arguments
$search_option = "default";
$machine_specs_file = "";
$prec = "";
$compiler_options_file = "";
$blocking_level = -1;

for ($i = 0; $i < scalar(@ARGV); $i++)
{
   if ($ARGV[$i] eq "-long")
   {
      $search_option = "long";
   }
   elsif ($ARGV[$i] eq "-default")
   {
      $search_option = "default";
   }
   elsif ($ARGV[$i] eq "-short")
   {
      $search_option = "short";
   }
   elsif ($ARGV[$i] eq "-machine")
   {
      $machine_specs_file = $ARGV[$i+1];
      $i++;
   }
   elsif ($ARGV[$i] eq "-prec")
   {
      $prec = $ARGV[$i+1];
      $i++;
   }
   elsif ($ARGV[$i] eq "-ccopt")
   {
      $compiler_options_file = $ARGV[$i+1];
      $i++;
   }
   elsif ($ARGV[$i] eq "-level")
   {
      $blocking_level = $ARGV[$i+1];
      $i++;
   }
   else
   {
      warn "***Ignoring unknown option " . $ARGV[$i] ."\n";
   }
}


if (($machine_specs_file eq "") ||
    ($prec eq "") ||
    ($compiler_options_file eq "") ||
    ($blocking_level < 0))
{
   die "Useage:\n\t$useage\n\n";
}

if (!(-e $compiler_options_file))
{
   die "***Compiler options file " . $compiler_options_file . " does not exist.\n\n";
}

if (($blocking_level > 2) || ($blocking_level < 0))
{
   die "***Blocking level " . $blocking_level . " not supported.\n\n";
}

if (($search_option ne "short") && ($search_option ne "default") &&
    ($search_option ne "long"))
{
   die "***Search option must be either \"long\", \"default\" or \"short\".\n\n";
}

if (($prec ne "single") && ($prec ne "double"))
{
   die "***Precision must be either \"single\" or \"double\".\n\n";
}


## Read search parameters.
require "search_readparams.pl";
&read_machine_specs($machine_specs_file);
&read_l0_search_params("$path_PHiPAC/$dir_SEARCH/opt_$search_option.L0_search");

## Start search procedure by installing IPM/RPRF

$cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_start.pl $machine_specs_file";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
if (!$NO_ACTION)
{
   system($cc_call);
   die "***Can't initialize search directory.\n\n"
      if ($? != 0);
}


## L0 core search

$cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_L0core.pl $machine_specs_file $prec $compiler_options_file $path_PHiPAC/$dir_SEARCH/opt_$search_option.L0_search";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
if (!$NO_ACTION)
{
   system($cc_call);
   die "***Can't perform L0 Core Search.\n\n"
      if ($? != 0);
}


## L0 performance sort

$cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_sort.pl L0CORE $prec";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
if (!$NO_ACTION)
{
   system($cc_call);
   die "***Can't perform performance sort for L0CORE.\n\n"
      if ($? != 0);
}
$top_file = "perf/L0CORE_top.$prec";


if ($blocking_level > 0)
{
   ## L1 search

   $cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_L1.pl $machine_specs_file $top_file $path_PHiPAC/$dir_SEARCH/opt_$search_option.L1_search";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;
   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't perform L1 Search.\n\n"
	 if ($? != 0);
   }


   ## L1 performance sort

   $cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_sort.pl L1 $prec";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;
   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't perform performance sort for L1.\n\n"
	 if ($? != 0);
   }
   $top_file = "perf/L1_top.$prec";


   ## L0 general search

   if ($do_L0_gen_frng == 1)
   {
      $cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_L0gen.pl $machine_specs_file $top_file $path_PHiPAC/$dir_SEARCH/opt_$search_option.L0_search";
      print STDOUT ">>> $cc_call\n" if $verbosity > 5;
      if (!$NO_ACTION)
      {
	 system($cc_call);
	 die "***Can't perform L0GEN Search.\n\n"
	    if ($? != 0);
      }
      
      $cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_sort.pl L0GEN $prec";
      print STDOUT ">>> $cc_call\n" if $verbosity > 5;
      if (!$NO_ACTION)
      {
	 system($cc_call);
	 die "***Can't perform performance sort for L0GEN.\n\n"
	    if ($? != 0);
      }
      $top_file = "perf/L0GEN_top.$prec";
   }


   if ($blocking_level > 1)
   {
      ## L2 search

      $cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_L2.pl $machine_specs_file $top_file $path_PHiPAC/$dir_SEARCH/opt_$search_option.L2_search";
      print STDOUT ">>> $cc_call\n" if $verbosity > 5;
      if (!$NO_ACTION)
      {
	 system($cc_call);
	 die "***Can't perform L2 Search.\n\n"
	    if ($? != 0);
      }


      ## L2 performance sort

      $cc_call = "perl $path_PHiPAC/$dir_SEARCH/search_sort.pl L2 $prec";
      print STDOUT ">>> $cc_call\n" if $verbosity > 5;
      if (!$NO_ACTION)
      {
	 system($cc_call);
	 die "***Can't perform performance sort for L2.\n\n"
	    if ($? != 0);
      }
      $top_file = "perf/L2_top.$prec";
   }
}


## Make GEMM libraries

$cc_call = "perl $path_PHiPAC/$dir_SEARCH/make_gemm.pl $machine_specs_file $prec $top_file";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
if (!$NO_ACTION)
{
   system($cc_call);
   die "***Can't create GEMM library.\n\n"
      if ($? != 0);
}


exit(0);

